#include "Data.h"


Node::Node(int p_node_id, const std::array<double, 3>& p_xyz, bool p_flag_vertex)
        : node_id(p_node_id)
        , xyz(p_xyz)
        , flag_vertex(p_flag_vertex)
{}

Node::Node(const Node& p_node)
        : node_id(p_node.node_id)
        , xyz(p_node.xyz)
        , flag_vertex(p_node.flag_vertex)
{}


std::istream& operator>>(std::istream& p_in, Node& p_node) {
    p_in >> p_node.node_id;
    p_in >> p_node.xyz.at(0)
         >> p_node.xyz.at(1)
         >> p_node.xyz.at(2);
    p_in >> p_node.flag_vertex;
    return p_in;
}

std::ostream& operator<<(std::ostream& p_out, const Node& p_node) {
    p_out << p_node.node_id << std::endl;
    p_out << p_node.xyz.at(0) << " "
          << p_node.xyz.at(1) << " "
          << p_node.xyz.at(2);
    p_out << p_node.flag_vertex << std::endl;
    return p_out;
}

//Finite element
FiniteElement::FiniteElement(int p_elem_id, int p_material_id, const std::vector<int>& p_node_id)
        : f_elem_id(p_elem_id)
        , material_id(p_material_id)
        , nodes_id(p_node_id)
{}

std::istream& operator>>(std::istream& p_in, FiniteElement& p_f_elem) {
    p_in >> p_f_elem.f_elem_id;
    p_in >> p_f_elem.material_id;
    p_in >> p_f_elem.nodes_id[0] >> p_f_elem.nodes_id[1] >> p_f_elem.nodes_id[2] >> p_f_elem.nodes_id[3];
    return p_in;
}

std::ostream& operator<<(std::ostream& p_out, const FiniteElement& p_f_elem)
{
    p_out << p_f_elem.f_elem_id << std::endl;
    p_out << p_f_elem.material_id << std::endl;
    p_out << "[ " << p_f_elem.nodes_id[0]
          << " " << p_f_elem.nodes_id[1]
          << " " << p_f_elem.nodes_id[2]
          << " " << p_f_elem.nodes_id[3]
          << " ]";
    return p_out;
}

bool FiniteElement::operator<(const FiniteElement& p_el) const {
    return f_elem_id < p_el.f_elem_id;
}

//Boundary finite element
BoundaryFiniteElement::BoundaryFiniteElement(int p_elem_id, int p_bound_id, const std::vector<int>& p_node_id)
        : bf_elem_id(p_elem_id)
        , boundary_id(p_bound_id)
        , nodes_id(p_node_id)
{}

std::istream& operator>>(std::istream& p_in, BoundaryFiniteElement& p_bf_elem) {
    p_in >> p_bf_elem.bf_elem_id;
    p_in >> p_bf_elem.boundary_id;
    p_in >> p_bf_elem.nodes_id[0] >> p_bf_elem.nodes_id[1] >> p_bf_elem.nodes_id[2];
    return p_in;
}

std::ostream& operator<<(std::ostream& p_out, const BoundaryFiniteElement& p_bf_elem) {
    p_out << p_bf_elem.bf_elem_id << std::endl;
    p_out << p_bf_elem.boundary_id << std::endl;
    p_out << "[ " << p_bf_elem.nodes_id[0]
          << " " << p_bf_elem.nodes_id[1]
          << " " << p_bf_elem.nodes_id[2]
          << " ]";
    return p_out;
}


//Edge
Edge::Edge(int p_id_first_node, int p_id_second_node)
        :nodes_of_edge(std::make_pair(p_id_first_node, p_id_second_node))
{}

void Edge::update_mid(int p_middle_node) {
    middle_node = p_middle_node;
}

Edge& Edge::operator=(const Edge& p_edge) {
    if (this != &p_edge) {
        this->nodes_of_edge = p_edge.nodes_of_edge;
        this->middle_node = p_edge.middle_node;
    }
    return *this;
}

bool Edge::operator==(const Edge& p_edge) const {
    return (nodes_of_edge.first == p_edge.nodes_of_edge.first) &&
           (nodes_of_edge.second == p_edge.nodes_of_edge.second);
}

std::istream& operator>>(std::istream& p_in, Edge& p_edge) {
    p_in >> p_edge.nodes_of_edge.first
         >> p_edge.nodes_of_edge.second;
    return p_in;
}

std::ostream& operator<<(std::ostream& p_out, Edge& p_edge) {
    p_out << "[" << p_edge.nodes_of_edge.first
          << ", " << p_edge.nodes_of_edge.second
          << "]";
    return p_out;
}