#ifndef MeshLoader_h
#define MeshLoader_h

#include "Data.h"
#include <set>
#include <vector>
#include <ranges>
#include <fstream>
#include <algorithm>



class MeshLoader{
    Node get_middle_node(const Edge&);
    std::array<double, Node::dimension> get_middle_of_edge(const Edge&);
    Node& get_node_by_id(int);
    std::array<double, Node::dimension> get_xyz_of_node_by_id(int);
protected:
    std::vector<Node> node;
    std::vector<FiniteElement> f_elem;
    std::vector<BoundaryFiniteElement> bf_elem;
public:
    virtual void loadMesh(const std::string&) = 0;

    virtual ~MeshLoader() = default;
    const std::vector<Node>& get_nodes() const;
    const std::vector<FiniteElement>& get_f_elems() const;

    const std::vector<BoundaryFiniteElement>& Get_bf_elems() const;
    std::vector<int> find_f_elem_by_nodes(int id1, int id2, int id3) const;
    std::vector<int> find_f_elem_by_edge(int id1, int id2) const;
    std::set<int> get_nodes_by_boundary(int id) const;
    std::set<int> get_f_elems_by_material(int id) const;

    std::set<int> get_bf_elems_by_boundary(int id) const;
    void insertion_in_middle();

    std::vector<std::set<int>> get_neighbor_nodes() const;
    void print_node(const Node &node);
    void print_f_elem(const FiniteElement &f_elem);
    void print_bf_elem(const BoundaryFiniteElement &bf_elem);


    void printAngles(const std::string& filepath);
    const std::array<double, 3>& findNodeRanges(int id);
};
#endif